@@ -1,4 +1,4 @@
-# Copyright 2012-2017 The Meson development team
+# Copyright 2012-2022 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -59,7 +59,7 @@ class StaticLinker:
     def get_exelist(self) -> T.List[str]:
         return self.exelist.copy()
 
-    def get_std_link_args(self, is_thin: bool) -> T.List[str]:
+    def get_std_link_args(self, env: 'Environment', is_thin: bool) -> T.List[str]:
         return []
 
     def get_buildtype_linker_args(self, buildtype: str) -> T.List[str]:
@@ -176,7 +176,7 @@ class ArLikeLinker(StaticLinker):
         # in fact, only the 'ar' id can
         return False
 
-    def get_std_link_args(self, is_thin: bool) -> T.List[str]:
+    def get_std_link_args(self, env: 'Environment', is_thin: bool) -> T.List[str]:
         return self.std_args
 
     def get_output_args(self, target: str) -> T.List[str]:
@@ -189,7 +189,7 @@ class ArLikeLinker(StaticLinker):
 class ArLinker(ArLikeLinker):
     id = 'ar'
 
-    def __init__(self, exelist: T.List[str]):
+    def __init__(self, for_machine: mesonlib.MachineChoice, exelist: T.List[str]):
         super().__init__(exelist)
         stdo = mesonlib.Popen_safe(self.exelist + ['-h'])[1]
         # Enable deterministic builds if they are available.
@@ -202,18 +202,26 @@ class ArLinker(ArLikeLinker):
         self.std_args = [stdargs]
         self.std_thin_args = [stdargs + thinargs]
         self.can_rsp = '@<' in stdo
+        self.for_machine = for_machine
 
     def can_linker_accept_rsp(self) -> bool:
         return self.can_rsp
 
-    def get_std_link_args(self, is_thin: bool) -> T.List[str]:
+    def get_std_link_args(self, env: 'Environment', is_thin: bool) -> T.List[str]:
         # FIXME: osx ld rejects this: "file built for unknown-unsupported file format"
-        if is_thin and not mesonlib.is_osx():
+        if is_thin and not env.machines[self.for_machine].is_darwin():
             return self.std_thin_args
         else:
             return self.std_args
 
 
+class AppleArLinker(ArLinker):
+
+    # mostly this is used to determine that we need to call ranlib
+
+    id = 'applear'
+
+
 class ArmarLinker(ArLikeLinker):
     id = 'armar'
 
@@ -225,7 +233,7 @@ class DLinker(StaticLinker):
         self.arch = arch
         self.__rsp_syntax = rsp_syntax
 
-    def get_std_link_args(self, is_thin: bool) -> T.List[str]:
+    def get_std_link_args(self, env: 'Environment', is_thin: bool) -> T.List[str]:
         return ['-lib']
 
     def get_output_args(self, target: str) -> T.List[str]:
@@ -700,14 +708,32 @@ class GnuLikeDynamicLinkerMixin:
         return (args, rpath_dirs_to_remove)
 
     def get_win_subsystem_args(self, value: str) -> T.List[str]:
-        if 'windows' in value:
-            args = ['--subsystem,windows']
-        elif 'console' in value:
-            args = ['--subsystem,console']
-        else:
-            raise MesonException(f'Only "windows" and "console" are supported for win_subsystem with MinGW, not "{value}".')
+        # MinGW only directly supports a couple of the possible
+        # PE application types. The raw integer works as an argument
+        # as well, and is always accepted, so we manually map the
+        # other types here. List of all types:
+        # https://github.com/wine-mirror/wine/blob/3ded60bd1654dc689d24a23305f4a93acce3a6f2/include/winnt.h#L2492-L2507
+        subsystems = {
+            "native": "1",
+            "windows": "windows",
+            "console": "console",
+            "posix": "7",
+            "efi_application": "10",
+            "efi_boot_service_driver": "11",
+            "efi_runtime_driver": "12",
+            "efi_rom": "13",
+            "boot_application": "16",
+        }
+        versionsuffix = None
         if ',' in value:
-            args[-1] = args[-1] + ':' + value.split(',')[1]
+            value, versionsuffix = value.split(',', 1)
+        newvalue = subsystems.get(value)
+        if newvalue is not None:
+            if versionsuffix is not None:
+                newvalue += f':{versionsuffix}'
+            args = [f'--subsystem,{newvalue}']
+        else:
+            raise mesonlib.MesonBugException(f'win_subsystem: {value!r} not handled in MinGW linker. This should not be possible.')
 
         return self._apply_prefix(args)
 
@@ -806,6 +832,11 @@ class GnuBFDDynamicLinker(GnuDynamicLinker):
     id = 'ld.bfd'
 
 
+class MoldDynamicLinker(GnuDynamicLinker):
+
+    id = 'ld.mold'
+
+
 class LLVMDynamicLinker(GnuLikeDynamicLinkerMixin, PosixDynamicLinkerMixin, DynamicLinker):
 
     """Representation of LLVM's ld.lld linker.
@@ -1136,7 +1167,7 @@ class PGIStaticLinker(StaticLinker):
         self.id = 'ar'
         self.std_args = ['-r']
 
-    def get_std_link_args(self, is_thin: bool) -> T.List[str]:
+    def get_std_link_args(self, env: 'Environment', is_thin: bool) -> T.List[str]:
         return self.std_args
 
     def get_output_args(self, target: str) -> T.List[str]:
